# $Id: Repsys.pm 258595 2009-07-25 16:41:57Z guillomovitch $

package MDV::Repsys;

use strict;
use warnings;
use Carp;
use Config::IniFiles;
use Date::Parse;
use Date::Format;
use File::Temp qw(tempdir);
use File::Path;
use File::Spec;
use POSIX qw(getcwd setlocale LC_ALL);
use RPM4;

use SVN::Client;
use SVN::Core;
use UNIVERSAL::require;
use version; our $VERSION = qv('2.0.6');

use constant {
    DEFAULT_VERSION  => 'current',
    DEFAULT_REVISION => 'HEAD',
    TEMPDIR_TEMPLATE => 'mdvsys_XXXX'
};

=head1 NAME

MDV::Repsys

=head1 DESCRIPTION

Perl interface with Mandriva build system

=head1 AUTHORS

Olivier Thauvin <nanardon@mandriva.org>

Guillaume Rousse <guillomovitch@mandriva.org>

Michaël Scherer <misc@mandriva.org>

=cut

my %rpm_dirs = (
    sourcedir => 'SOURCES',
    patchdir  => 'SOURCES',
    specdir   => 'SPECS',
    builddir  => 'BUILD',
    rpmdir    => 'RPMS',
    srcrpmdir => 'SRPMS',
);

setlocale(LC_ALL, "C");

=head1 CLASS METHODS

=head2 new(%options)

Create a new MDV::Repsys object

Available options:

=over

=item configfile

Use this repsys configuration file instead /etc/repsys.conf

=item commit true/false

Allow commit action (default: true)

=item submit true/fals

Allow submit action (default: true).

=back

=cut

sub new {
    my $class = shift;
    my %options = (
        configfile => '/etc/repsys.conf',
        verbosity  => 0,
        commit     => 1,
        submit     => 1,
        @_
    );

    my $config = Config::IniFiles->new(
        -file => $options{configfile}
    );
    croak "Unable to read configuration file $options{configfile}\n" 
        unless $config;

    # override system configuration with local configuration if present
    my $homerepsys = (
        $ENV{REPSYS_CONF} ?
        $ENV{REPSYS_CONF} :
        "$ENV{HOME}/.repsys/repsys.conf"
    );

    if (-r $homerepsys) {
        $config = Config::IniFiles->new(
            -file   => $homerepsys, 
            -import => $config
        );
    }

    my $top_url = $config->val('global', 'default_parent');
    croak "No default_parent parameter defined" unless $top_url;
    $top_url =~ s:/+[^/]+/*$::;

    my $mirror_url = $config->val('global', 'mirror');
    $mirror_url =~ s:/+[^/]+/*$:: if $mirror_url;

    my $self = bless {
        config    => $config,
        pool      => SVN::Pool->new_default(),
        svn       => SVN::Client->new(),
        top_url   => $top_url,
        mirror_url => $mirror_url,
        commit    => $options{commit},
        submit    => $options{submit} && $options{commit} ,
        verbosity => $options{verbosity},
    }, $class;

    return $self;
}

=head1 INSTANCE METHODS

=head2 set_verbosity($level)

Set verbosity to given level.

=cut

sub set_verbosity {
    my ($self, $level) = @_;

    $self->{verbosity} = $level;
}

=head2 unpack_package($rpmfile, $destdir)

Unpack a source package into a directory.
Return corresponding spec file.

=cut

sub unpack_package {
    my ($self, $rpmfile, $destdir) = @_;

    $self->_print_msg(3, 'Unpacking package %s', $rpmfile);

    if (! -d $destdir) {
        mkdir $destdir or croak "Can't create $destdir: $!";
    }

    $self->_set_rpm_dirs_from_topdir($destdir);

    # make rpm output informations starting from log level 4
    RPM4::setverbosity($self->{verbosity} >= 4 ? $self->{verbosity} + 2 : 0);
    my ($specfile) = RPM4::installsrpm($rpmfile);
    RPM4::setverbosity(0);

    return $specfile;
}

=head2 find_current_sources($dir)

Return the list of sources (meaning: sources, icons and spec files) from the
package found in given directory, according to their usage in the spec file.

=cut

sub find_current_sources {
    my ($self, $workdir) = @_;

    $self->_set_rpm_dirs_from_topdir($workdir);
    my $specfile = (glob(RPM4::expand('%_specdir/*.spec')))[0]
        or croak "Can't find any spec file\n";
    $self->_print_msg(3, 'Looking sources from spec file %s', $specfile);
   
    my $spec = RPM4::Spec->new($specfile, anyarch => 1, force => 0)
        or croak "Can't parse spec file $specfile\n";

    my @sources = ( $specfile ); # specfile should be first !
    my $sourcedir = RPM4::expand('%_sourcedir');
    foreach my $file ($spec->sources(), $spec->icon()) {
        push(@sources, "$sourcedir/$file");
    }

    return @sources;
}

=head2 find_unsync_sources($dir, @sources)

Return two lists of files that should be added or removed from the svn
repository according to their current status.

=cut

sub find_unsync_sources {
    my ($self, $workdir, @sources) = @_;

    my %sources = map { $_ => 1 } @sources;

    my @dirs = qw/SPECS SOURCES/;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    # ensure spec and source directory are already versionned
    my $children = $self->{svn}->ls($workdir, 'HEAD', 0, $pool);
    foreach my $dir (@dirs) {
        next if $children->{$dir};
        $self->_print_msg(4, "Adding %s", $dir);
        $self->{svn}->add("$workdir/$dir", 0, $pool);
    }

    my (@needadd, @needdel);
    foreach my $dir (@dirs) {
        $self->{svn}->status(
            "$workdir/$dir",
            'HEAD',
            sub {
                my ($entry, $status) = @_;
                # skip directory itself
                return if $entry eq "$workdir/$dir";

                my $text_status = $status->text_status();
                if ($text_status == $SVN::Wc::Status::unversioned) {
                    push(@needadd, $entry) if $sources{$entry};
                } elsif (
                    $text_status == $SVN::Wc::Status::normal   ||
                    $text_status == $SVN::Wc::Status::added    ||
                    $text_status == $SVN::Wc::Status::modified
                ) {
                    push(@needdel, $entry) if !$sources{$entry};
                }
            },
            1, # recursive
            1, # get_all
            0, # update
            1, # no_ignore
            $pool
        );
    }
    
    return(\@needadd, \@needdel);
}

=head2 sync_sources($needadd, $needdel)

Add and remove sources from the repository.

=cut

sub sync_sources {
    my ($self, $needadd, $needdel) = @_;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    foreach my $toadd (sort @{$needadd || []}) {
        $self->_print_msg(4, "Adding %s", $toadd);
        $self->{svn}->add($toadd, 0, $pool);
    }
    foreach my $todel (sort @{$needdel || []}) {
        $self->_print_msg(4, "Removing %s", $todel);
        $self->{svn}->delete($todel, 1, $pool);
    }
}

=head2 build_package($specfile, %options)

Build package from given spec file. Return the source package, followed by the
list of binary packages.

Available options:

=over

=item binaries => true/false (default: true)

build binaries packages

=item source => true/false (default: true)

build source package

=item binaries_destdir $dir

create binary packages in given directory.

=item source_destdir $dir

create source package in given directory.

=item define => values

Pass each given value to underlying command, using --define switch.

=back

=cut

sub build_package {
    my ($self, $specfile, %options) = @_;

    $self->_print_msg(3, 'Building package from spec file %s', $specfile);

    # compute top directory
    $specfile = File::Spec->rel2abs($specfile);
    my ($topdir) = $specfile =~ m|^(.*)/[^/]+/[^/]+\.spec$|;

    my @results;
    require Youri::Package::RPM::Builder;
    my $builder = Youri::Package::RPM::Builder->new(
        topdir                 => $topdir,
        build_requires_command =>
            $self->{config}->val('helper', 'install-buildrequires') || '',
        build_results_callback => sub { @results = @_ },
        verbose                =>
            ($self->{verbosity} > 2 ? $self->{verbosity} - 2 : 0)
    );

    my $rpm_options = " --define '_signature \"\"'";
    $rpm_options .= " --define '_rpmdir $options{binaries_destdir}'"
        if $options{binaries_destdir};
    $rpm_options .= " --define '_srcrpmdir $options{source_destdir}'"
        if $options{source_destdir};
    $rpm_options .= " --define '$_'" foreach
        $options{define} ? @{$options{define}} : ();

    eval {
        $builder->build(
            $specfile,
            build_binaries =>
                defined $options{binaries} ?  $options{binaries} : 1,
            build_sources =>
                defined $options{sources}  ?  $options{sources} : 1,
            rpm_options => $rpm_options
        );
    }; 
    if ($@) {
        $self->_print_msg(2, 'Keeping build directory in %s', $topdir)
            if $options{'keep-on-failure'};
        croak "Can't build package: $@\n";
    }

    return @results;
}

=head2 get_packages_top_url()

Return the top-level svn url for packages

=cut

sub get_packages_top_url {
    my ($self, %options) = @_;
    # return preferentialy mirror url, if defined and possible
    return $options{write}     ? $self->{top_url}    :
           $self->{mirror_url} ? $self->{mirror_url} :
                                 $self->{top_url}    ;
}

=head2 get_package_url($package, %options)

Return the svn url for package named $package

=cut

sub get_package_url {
    my ($self, $package, %options) = @_;
    return sprintf(
        '%s/%s',
        $self->get_package_parent_url($package, %options),
        $options{version} || DEFAULT_VERSION
    );
}

=head2 get_package_parent_url($package, %options)

Return the svn parent url for package named $package

=cut

sub get_package_parent_url {
    my ($self, $package, %options) = @_;

    my $distribution = $options{distribution} || 'cooker';
    my $branch       = $options{branch};
    my $url =
        $self->get_packages_top_url(%options)                              .
        ($distribution eq 'cooker' ? '/cooker' : "/updates/$distribution") .
        "/$package"                                                        .
        ($branch ? "/branches/$branch" : '');

    return $url;
}

=head2 get_package_name_from_directory($dir)

Return the package name locate in the working directory B<$dir>

=cut

=head2 get_package_name_from_current_directory

Return the package name from current directory

=cut

sub get_package_name_from_current_directory { 
    my ($self) = @_;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    my $url;
    my $receiver = sub {
        my (undef, $info, undef) = @_;
        $url = $info->URL();
    };
    eval { 
        $self->{svn}->info(getcwd(), undef, 'WORKING', $receiver, 0, $pool);
    };
    if ($@) {
        return;
    }
    my $parent = $self->{config}->val('global', 'default_parent');
    return unless $parent;

    $url =~ /^\Q$parent\E\/([^\/]*)\/?.*$/;
    my $name = $1;

    return $name;
}

=head2 checkout_package($package, $destdir, %options)

Checkout a package from svn into $destdir directory

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub checkout_package {
    my ($self, $package, $destdir, %options) = @_;

    $destdir ||= $package;

    $self->_print_msg(3,
        'Checking out package %s into directory %s', $package, $destdir
    );

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    my $revision;
    eval {
        $revision = $self->{svn}->checkout(
            $self->get_package_url($package, %options),
            $destdir,
            $options{revision} || DEFAULT_REVISION,
            1,
            $pool
        );
    };
    if ($@) {
        if ($@ =~ /^Bad URL passed to RA layer/) {
            croak "No such package $package\n";
        } else {
            croak "Can't checkout $package: $@\n";
        }
    }

    return $revision;
}

=head2 print_raw_package_changelog($package, $handle, %options)

Print the raw subversion log for package to $handle.
If not specified, $handle is set to STDOUT.

=cut

sub print_raw_package_changelog {
    my ($self, $package, $handle, %options) = @_;

    $handle ||= \*STDOUT;

    $self->_print_msg(3, 'Printing raw changelog for package %s', $package);

    foreach my $entry ($self->_get_package_changelog($package, %options)) {
        print $handle $self->_format_changelog_entry($entry);
    }
}

=head2 print_package_changelog($package, $handle, %options)

Print the rpm-formated log for package to $handle.
If not specified, $handle is set to STDOUT.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub print_package_changelog {
    my ($self, $package, $handle, %options) = @_;

    $handle ||= \*STDOUT;

    $self->_print_msg(3, 'Printing changelog for package %s', $package);
    my @entries = (
        $self->_get_package_changelog($package, %options),
        $self->_get_old_package_changelog($package, %options)
    );
 
    print $handle "\%changelog\n";

    foreach my $entry (sort {
            $b->{time} &&  $a->{time} ?
            $b->{time} <=> $a->{time} :
            $a->{time} <=> $b->{time}
        } grep { $_ } @entries) {
        print $handle $self->_format_changelog_entry($entry);
    }
}

=head2 print_old_package_changelog($package, $handle, %options)

Read old changelog entry from svn and write it into $handle.
If not specified, $handle is set to STDOUT.

=cut

sub print_old_package_changelog {
    my ($self, $package, $handle, %options) = @_;
    
    $handle ||= \*STDOUT;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    $self->_print_msg(3, 'Printing old changelog for package %s', $package);
    eval {
        $self->{svn}->cat(
            $handle,
            sprintf(
                "%s/%s/log",
                $self->{config}->val('log', 'oldurl'),
                $package,
            ),
            $options{revision} || DEFAULT_REVISION,
            $pool
        ); 
    };
    if ($@) {
        if ($@ =~ /^Filesystem has no item/) {
            # no changelog for this package
        } else {
            croak "Can't get old changelog for $package: $@\n";
        }
    }
}

=head2 strip_changelog_from_spec($specfile)

Available options:

=over

=item target => $filename

write new spec file as given filename

=back

=cut

sub strip_changelog_from_spec {
    my ($self, $specfile, %options) = @_;
    
    $self->_print_msg(3, 'Stripping changelog from spec file %s', $specfile);

    my $new_specfile = $options{target} ? $options{target} : $specfile . '.new';
    open(my $in, '<', $specfile)
        or croak "Can't open $specfile for reading: $!";
    open(my $out, '>', $new_specfile)
        or croak "Can't open $new_specfile for writing: $!";

    while (my $line = <$in>) {
        last if $line =~ /^\%changelog/;
        print $out $line;
    }

    my $changelog;
    while (my $line = <$in>) {
        next if !$line && !$changelog;
        $changelog .= $line;
    }

    close($out);
    close($in);

    if (!$options{target}) {
        rename($new_specfile, $specfile)
            or croak "Can't move $new_specfile to $specfile: $!";
    };

    return $changelog;
}

=head2 add_changelog_to_spec($specfile, %options)

Rewrite a spec file, adding changelog from repository.

Available options:

=over

=item target => $filename

write new spec file as given filename

=back

=cut

sub add_changelog_to_spec {
    my ($self, $specfile, %options) = @_;

    $self->_print_msg(3, 'Adding changelog to spec file %s', $specfile);

    $self->_set_rpm_dirs_from_specfile($specfile);
    my $spec = RPM4::Spec->new($specfile, anyarch => 1, force => 1)
        or croak "Can't parse specfile $specfile\n";
    my $header = $spec->srcheader();
    my $package = $header->queryformat('%{NAME}');

    my $new_specfile = $options{target} ? $options{target} : $specfile . '.new';
    open(my $in, '<', $specfile)
        or croak "Can't open $specfile for reading: $!";
    open(my $out, '>', $new_specfile)
        or croak "Can't open $new_specfile for writing: $!";

     while (my $line = <$in>) {
        print $out $line;
    }

    $self->print_package_changelog(
        $package,
        $out,
        %options,
    );

    close($out);
    close($in);

    if (!$options{target}) {
        rename($new_specfile, $specfile)
            or croak "Can't move $new_specfile to $specfile: $!";
    };

}

=head2 get_last_package_revision($package, %options)

Return the last revision change for a package.

=cut

sub get_last_package_revision {
    my ($self, $package, %options) = @_;
    my $url = $self->get_package_url($package, %options);
    my $leafs;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $leafs = $self->{svn}->ls(
            $url, 
            $options{revision} || DEFAULT_REVISION,
            1,
            $pool
        );
    };
    if ($@) {
        croak "Can't get information from $url: $@\n";
    }
    my $revision = 0;
    foreach my $leaf (%{$leafs || {}}) {
        defined($leafs->{$leaf}) or next;
        if ($leafs->{$leaf}->created_rev > $revision) {
            $revision = $leafs->{$leaf}->created_rev;
        }
    }
        
    return $revision;
}

=head2 get_last_directory_revision($dir, %options)

Return the real last revision change for package checkout into $dir.

=cut

sub get_last_directory_revision {
    my ($self, $dir, %options) = @_;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    $self->_print_msg(4, 'Finding last revision from directory %s', $dir);
    my $revision = 0;
    eval {
        $self->{svn}->status(
            $dir,
            $options{revision} || DEFAULT_REVISION,
            sub {
                my ($path, $status) = @_;
                my $entry = $status->entry();
                return unless $entry;
                my $entry_revision = $entry->cmt_rev();
                $revision = $entry_revision if $revision < $entry_revision;
            },
            1, # recursive
            1, # get_all
            0, # update
            0, # no_ignore
        );
    };
    if ($@) {
        croak "can't get status of $dir: $@\n";
    }

    return $revision
}

=head2 extract_package($package, %options)

Extract a source package from the subversion repository into the current
directory. Return the package file.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=item destdir => $dir

Use given directory as destination.

=item prefix => true/false

Prefix the package file by revision number, so as to be compatible with
mandriva submission tool.

=back

=cut

sub extract_package {
    my ($self, $package, %options) = @_;

    $self->_print_msg(2, 'Extracting package %s', $package);

    my $dir = tempdir(
        TEMPDIR_TEMPLATE,
        TMPDIR => 1,
        CLEANUP => ($options{'keep-on-failure'} ? 0 : 1)
    );

    $self->checkout_package($package, $dir, %options);

    my $specfile = "$dir/SPECS/$package.spec";
    
    $self->add_changelog_to_spec(
        $specfile,
        %options,
        package => $package
    );

    my $destdir = $options{destdir} || getcwd();
    my ($file) = $self->build_package(
        $specfile,
        deps           => 0,
        binaries       => 0,
        source_destdir => $destdir
    );

    if ($options{prefix}) {
        my $revision = $self->get_last_directory_revision($dir, %options);
        my $new_file = '@' . $revision . (File::Spec->splitpath($file))[2];
        $self->_print_msg(3, 'Renaming %s to %s', $file, $new_file);
        rename $file, $new_file;
        $file = $new_file;
    }

    # remove temp directory immediatly, not at program termination
    rmtree($dir);

    return $file;
}

=head2 create_package_directory($package, %options)

Create a package directory on the svn.

=over

=item message => $message

Use given commit message (default: "create X", where X is package name).

=back

=cut

sub create_package_directory {
    my ($self, $package, %options) = @_;

    my $parent_url = $self->get_package_parent_url($package, %options);
    my $url = $self->get_package_url($package, %options);

    if ($self->_check_url_exists($url, %options)) {
        croak "$package is already inside svn\n";
    }

    $self->_print_msg(3, "Creating directory for package %s", $package);

    my @urls = (
        $parent_url,
        $url,
        "$url/SOURCES",
        "$url/SPECS",
    );

    if ($options{branch}) {
        my $branch_url = $self->get_package_parent_url(
            $package,
            %options, 
            branch => undef
        ) . '/branches';
        push(@urls, $branch_url)
            unless $self->_check_url_exists($branch_url);
    }
   
    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    my $message = $options{message} || "create $package";
    $self->{svn}->log_msg(sub { $_[0] = \$message; return 0; });
    eval { $self->{svn}->mkdir(\@urls, $pool); };
    $self->{svn}->log_msg(undef);
    if ($@) {
        croak "Can't create $package: $@\n";
    }
}

=head2 import_package($rpmfile, %options)

Import a source package into the svn.

=over

=item message => $message

Use given commit message (default: "import X", where X is package name)

=back

=cut

sub import_package {
    my ($self, $rpmfile, %options) = @_;

    my $header = RPM4::Header->new($rpmfile)
        or croak "Can't read rpm file $rpmfile\n";
    if ($header->hastag('SOURCERPM')) {
        croak "$rpmfile is not a source package\n";
    }
    my $package = $header->queryformat('%{NAME}');

    my $parent_url = $self->{config}->val('global', 'default_parent')
        or croak "No 'default_parent' defined in configuration";

    croak "$package is already inside svn\n"
        if $self->_check_url_exists("$parent_url/$package");

    $self->_print_msg(2, "Importing package %s", $package);
    my $dir = tempdir(TEMPDIR_TEMPLATE, TMPDIR => 1, CLEANUP => 1);
    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $self->{svn}->checkout(
            $parent_url,
            $dir,
            'HEAD', # What else ??
            0, # Don't be recursive !!
            $pool
        );
    };
    if ($@) {
        croak "Can't checkout $parent_url: $@\n";
    }

    my $pkgdir = "$dir/$package";

    $self->{svn}->mkdir($pkgdir);

    my $specfile = $self->unpack_package(
        $rpmfile,
        "$pkgdir/current",
    );

    $self->{svn}->add("$pkgdir/current", 1);

    my $changelog = $self->strip_changelog_from_spec(
        $specfile, 
        %options,
        package => $package,
    );

    $self->import_changelog(
        $changelog, 
        $package,
        %options
    ) if $changelog;
   
    my $release = $header->queryformat('%{VERSION}-%{RELEASE}');

    my $message = $options{message} || "import $package";

    $self->commit_directory(
        $pkgdir,
        %options,
        package => $package,
        message => $message
    );

    # remove temp directory immediatly, not at program termination
    rmtree($dir);
}

=head2 commit_directory($dir, %options)

Commit a directory into the repository. Return revision number.

Available options:

=over

=item message => $message

Use given commit message  (default: launch interactive editor).

=back

=cut

sub commit_directory {
    my ($self, $dir, %options) = @_;

    return unless $self->{commit};

    my $package = $options{package} || $dir;

    $self->sync_sources(
        $self->find_unsync_sources(
            $dir, $self->find_current_sources($dir)
        )
    ) if $options{sync};

    $self->_print_msg(3, "Committing directory %s", $package);

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    # set log callback
    my $message = $options{message};
    if (!$message) {

        my $changes = '';
        my $find_changes = sub {
            my ($entry, $status) = @_;
            # skip directory entries
            return if -d $entry;

            my $text_status = $status->text_status();
            my $change =
                $text_status == $SVN::Wc::Status::modified ? "modified" :
                $text_status == $SVN::Wc::Status::added    ? "added"    :
                $text_status == $SVN::Wc::Status::deleted  ? "removed"  :
                                                             ""         ;
            $changes .= sprintf(
                "file %s %s\n", File::Spec->abs2rel($entry), $change
            ) if $change;
        };

        foreach my $subdir (qw/SPECS SOURCES/) {
            $self->{svn}->status(
                "$dir/$subdir",
                'HEAD',
                $find_changes,
                1, # recursive
                1, # get_all
                0, # update
                1, # no_ignore
                $pool
            );
        }

        $message = $self->_get_commit_message_from_editor($changes);
    }
    $self->{svn}->log_msg(
        sub {
            $_[0] = \$message;
            return 0;
        }
    );

    # switch to actual repository if needed
    if ($self->{mirror_url}) {
        $self->{svn}->switch(
            $dir, 
            $self->get_package_url($package, write => 1),
            'HEAD',
            1,
            $pool
        );
    }

    my $info = $self->{svn}->commit($dir, 0, $pool);
    my $revision = $info ? $info->revision() : undef;

    # reset log callback
    $self->{svn}->log_msg(undef);

    # switch back to mirror if needed
    if ($self->{mirror_url}) {
        $self->{svn}->switch(
            $dir, 
            $self->get_package_url($package),
            'HEAD',
            1,
            $pool
        );
    }

    return $revision;
}

=head2 import_changelog($changelog, $package, %options)

Import given changelog into the repository.

Available options:

=over

=item message => $message

Use given commit message (default: "import old changelog for X", where X is
package name)

=back

=cut

sub import_changelog {
    my ($self, $changelog, $package, %options) = @_;

    my $parent_url = $self->{config}->val('log', 'oldurl')
        or croak "No 'oldurl' defined in configuration";

    croak "$package changelog is already inside svn\n"
        if $self->_check_url_exists("$parent_url/$package");

    $self->_print_msg(3, "Importing changelog for package %s", $package);
    my $dir = tempdir(TEMPDIR_TEMPLATE, TMPDIR => 1, CLEANUP => 1);
    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $self->{svn}->checkout(
            $parent_url,
            $dir,
            'HEAD',
            0,
            $pool
        );
    };
    if ($@) {
        croak "Can't checkout $parent_url: $@\n";
    }

    my $pkgdir = "$dir/$package";

    $self->{svn}->mkdir("$pkgdir");

    open(my $out, ">", "$pkgdir/log")
        or croak "Can't open new log file\n";
    print $out $changelog;
    close($out);

    $self->{svn}->add("$pkgdir/log", 0);

    my $message = $options{message} || "import old changelog for $package";
    $self->commit_directory($dir, message => $message);

    # remove temp directory immediatly, not at program termination
    rmtree($dir);
}

=head2 tag_package($package, %options)

Tag a package into the repository directly.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item message => $message

Use given commit message (default: "tag EV-R", where E is epoch, V is version,
and R is release).

=back

=cut

sub tag_package {
    my ($self, $package, %options) = @_;

    $self->_print_msg(3, 'Tagging package %s', $package);

    my $file = File::Temp->new(CLEANUP => 1);
    my $url = $self->get_package_url($package) . "/SPECS/$package.spec";

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $self->{svn}->cat(
            $file,
            $url,
            $options{revision} || DEFAULT_REVISION,
            $pool
        );
    };
    if ($@) {
        croak "Can't get spec file from $url: $@\n";
    }

    close($file);

    my $spec = RPM4::Spec->new($file, anyarch => 1, force => 1)
        or croak "Can't parse spec file from $url\n";
    my $header = $spec->srcheader;

    my $ev = $header->queryformat('%|EPOCH?{%{EPOCH}:}:{}|%{VERSION}');
    my $re = $header->queryformat('%{RELEASE}');

    my $tagurl = $self->get_package_url($package, version => 'releases');
    my $pristineurl = $self->get_package_url($package, version => 'pristine');

    if (!$self->_check_url_exists($tagurl)) {
        $self->{svn}->mkdir($tagurl);
    }

    if (!$self->_check_url_exists("$tagurl/$ev")) {
        $self->{svn}->mkdir("$tagurl/$ev");
    }

    if ($self->_check_url_exists("$tagurl/$ev/$re")) {
        croak "$tagurl/$ev/$re already exists\n";
    }

    my $message = $options{message} || "tag release $ev-$re";
    $self->{svn}->log_msg(
        sub {
            $_[0] = \$message;
            return 0;
        }
    );
    $self->_print_msg(4, 'Tagging %s to %s/%s', $package, $ev, $re);
    $self->{svn}->copy(
        $self->get_package_url($package),
        $options{revision} || DEFAULT_REVISION,
        "$tagurl/$ev/$re",
    );
    eval {
        $self->{svn}->delete($pristineurl, 1);
    };
    $self->_print_msg(4, 'Tagging package %s to pristine', $package);
    $self->{svn}->copy(
        $self->get_package_url($package, write => 1),
        $options{revision} || DEFAULT_REVISION,
        $pristineurl
    );
    $self->{svn}->log_msg(undef);
}

=head2 get_package_info($package, %options)

Return an hashref containing useful information about given package:

=over

=item size

The size of the package (sum of files size)

=item last_rev

The revision of the last changed

=item last_author

The author of the last change

=item last_time

The time of last change

=back

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub get_package_info {
    my ($self, $package, %options) = @_;
    my $url = $self->get_package_url($package, %options);
    my $leaves;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $leaves = $self->{svn}->ls(
            $url, 
            $options{revision} || DEFAULT_REVISION,
            1,
            $pool
        );
    };
    if ($@) {
        if ($@ =~ /^Filesystem has no item/) {
            croak "No such package $package\n";
        } else {
            croak "Can't get information from $url: $@\n";
        }
    }
    my $info = {
        size        => 0,
        last_rev    => 0,
        last_time   => 0,
        last_author => ''
    };
    foreach my $leaf (%{$leaves || {}}) {
        defined($leaves->{$leaf}) or next;
        $info->{size} += $leaves->{$leaf}->size;
        if ($leaves->{$leaf}->created_rev > $info->{last_rev}) {
            $info->{last_rev} = $leaves->{$leaf}->created_rev();
            $info->{last_time} = $leaves->{$leaf}->time();
            $info->{last_author} = $leaves->{$leaf}->last_author();
        }
    }

    return $info;
}

=head2 submit_package($package, %options)

Submit the package on the build host.

Available options:

=over

=item revision => $revision

Work on a specific revision (default to last revision).

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=item target => $target

Use given target as submission target (default to repsys configuration value).

=item define => values

Pass each given value to underlying command, using --define switch.

=back

=cut

sub submit_package {
    my ($self, $package, %options) = @_;

    return unless $self->{submit};

    $self->_print_msg(3, 'Submitting package %s', $package);
    
    my $parent_url = $self->get_package_parent_url($package, %options);
    if (!$self->_check_url_exists($parent_url, %options)) {
        croak "$package is not in svn\n";
    }

    my $revision = $options{revision} ||
                   $self->get_package_info($package, %options)->{last_rev};
    my $target   = $options{target} || 
                   $self->{config}->val('submit', 'default');
    my $host       = $self->{config}->val('submit', 'host');
    my $createsrpm = $self->{config}->val('helper', 'create-srpm');
    
    # TODO we can also use xml-rpc, even if not implemented on the server side
    my @command = (
        'ssh',
        $host,
        $createsrpm,
        $parent_url,
        '-r', $revision, 
        '-t', $target,
        (map { ('--define' => $_) } @{$options{define} || []}),
    );

    my $result = system(@command);
    if ($result != 0) {
        croak("Error during submission: $?\n");
    }
}

=head2 update_package($package, $version, %options)

Update a package to a new version.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=item force-version => $version

Force given package version.

=item force-release => $release

Force given package release.

=item keep-on-failure => true/false

Keep temporary directory in case of failure

=item message => $message

Use given commit message (default: "update to new version X", where X is new
version if given, "rebuild" otherwise)

=item target => $target

Use given target as submission target.

=item nobuild => true/false

Skip local build of the package.

=back

=cut

sub update_package {
    my ($self, $package, $version, %options) = @_;

    $self->_print_msg(2,
        'Updating package %s' . ($version ? ' to %s' : ''), $package, $version
    );

    my $dir = $options{tempdir} || tempdir(
        TEMPDIR_TEMPLATE,
        TMPDIR => 1,
        CLEANUP => ($options{'keep-on-failure'} ? 0 : 1)
    );

    $self->checkout_package(
        $package,
        $dir,
        %options
    );

    # ensure a full second delay between checkout and spec update
    # otherwise svn doesn't see differences
    sleep(1);

    $self->_set_rpm_dirs_from_topdir($dir);

    my $spec_file =
        $options{specfile} ||
        (glob(RPM4::expand('%_specdir/*.spec')))[0];

    Youri::Package::RPM::Updater->require(qv('0.4.0')) or croak $@;
    my $updater = Youri::Package::RPM::Updater->new(
        sourcedir => "$dir/SOURCES",
        verbose                =>
            ($self->{verbosity} > 2 ? $self->{verbosity} - 2 : 0)
    );

    eval {
        $updater->update_from_spec(
            $spec_file,
            $version,
            ($options{'release'} ?
                (release => $options{'release'}) :
                ()
            ),
            ($options{'spec-line-expression'} ?
                (spec_line_expression => $options{'spec-line-expression'}) :
                ()
            ),
            update_changelog => 0,
            # don't check version for perl packages
            check_new_version => ($package =~ /^perl-/ ? 0 : 1)
        );
    }; 
    if ($@) {
        croak "Can't update package: $@\n";
    }

    $self->build_package($spec_file, %options) unless $options{nobuild};

    # check current svn status before syncing
    $self->sync_sources(
        $self->find_unsync_sources(
            $dir, $self->find_current_sources($dir)
        )
    );

    my $message = $options{message} ? $options{message}
                : $version          ? "update to new version $version"
                :                     "rebuild" 
                ;
    $self->commit_directory($dir, message => $message);

    $self->submit_package($package, %options);

    # remove temp directory immediatly, not at program termination
    rmtree($dir);
}

=head2 mass_update($url)

Use given URL as a list of available packages update, and call update package
foreach of them

Available options:

=over

=item include => $regexp

Only process package matching given regexp

=item exclude => $regexp

Don't process package matching given regexp

=item keep-on-failure => true/false

Keep temporary directories in case of failure

=back

=cut

sub mass_update {
    my ($self, $url, %options) = @_;
    croak 'No URL given' unless $url;

    LWP::UserAgent->require() or croak $@;
    HTML::TableExtract->require() or croak $@;

    $self->_print_msg(1, 'Mass updating from %s', $url);

    $self->_print_msg(2, 'Retrieving report');
    my $agent = LWP::UserAgent->new();
    my $response = $agent->get($url);

    croak 'Unreachable URL: ' . $response->code()
        unless $response->is_success();

    my $parser;
    my $content = $response->content();
    if ($content =~ m/<title>age (global|individual) report/) {
        $parser = HTML::TableExtract->new(
            headers => [
                'source package',
            ]
        );
    } elsif ($content =~ m/<title>updates (global|individual) report/) {
        $parser = HTML::TableExtract->new(
            headers => [
                'source package',
                'available'
            ]
        );
    } else {
        croak 'Unknown report type';
    }

    $self->_print_msg(2, 'Parsing report content');
    $parser->parse($content);
    my ($include, $exclude);
    $include = qr/$options{include}/ if $options{include};
    $exclude = qr/$options{exclude}/ if $options{exclude};

    my $dir = tempdir(TEMPDIR_TEMPLATE, TMPDIR => 1, CLEANUP => 0);
    my $error_occured = 0;
    foreach my $row ($parser->rows()) {
        my ($name, $version) = ($row->[0], $row->[1]);
        # skip additional lines for spanning rows
        next unless $name; 
        next if $exclude && $name =~ $exclude;
        next if $include && $name !~ $include;
        eval {
            $self->update_package(
                $name,
                $version,
                tempdir => "$dir/$name"
            );
        };
        if ($@) {
            warn "Update failed for $name" .
                ($version ? " $version" : "") .
                "\n";
            $error_occured = 1;
        }
    }

    # remove temp directory immediatly, not at program termination
    rmtree($dir) unless $options{'keep-on-failure'} && $error_occured;
}

# private methods

# Set rpm macros corresponding to build directories, either relatively to given
# top directory if relative path is given, or absolutly if absolute path is
# given.

sub _set_rpm_dirs_from_topdir {
    my ($self, $topdir, %other_dirs) = @_;

    $topdir = File::Spec->rel2abs($topdir);

    foreach my $name (keys %rpm_dirs) {
        my $path = $other_dirs{$name} || $rpm_dirs{$name};
        if ($path !~ /^\//) {
            $path = "$topdir/$path";
        }
        RPM4::add_macro("_$name $path");
    }
}

# Set rpm macros corresponding to build directories, calling
# _set_rpm_dirs_from_topdir with the parent directory of given spec file
# directory.

sub _set_rpm_dirs_from_specfile {
    my ($self, $specfile, %other_dirs) = @_;

    my ($topdir) = $specfile =~ /^(.*)\/.*\/.*$/;
    $self->_set_rpm_dirs_from_topdir($topdir, %other_dirs);
}

# Create rpm build directories

sub _create_rpm_dirs {
    my ($self, @dirs) = @_;

    @dirs = keys %rpm_dirs unless @dirs;

    foreach my $name (@dirs) {
        my $path = RPM4::expand("%_$name");
        if (! -d $path) {
            $self->_print_msg(5, 'Creating directory %s', $path);
            if (!mkdir($path)) {
                croak "can't create directory $path: $!\n";
            }
        }
    }
}

# Print a trace message, according to current log level

sub _print_msg {
    my ($self, $level, $fmt, @args) = @_;
    return if $level > $self->{verbosity};
    printf("$fmt\n", @args);
}

# Get a commit message using an external editor

sub _get_commit_message_from_editor {
    my ($self, $footer) = @_;

    my $delimiter = "--This line, and following ones, will be ignored--";
    
    my $tmp = File::Temp->new(CLEANUP => 1);
    print $tmp "\n";
    print $tmp $delimiter . "\n";
    print $tmp $footer if $footer;
    close($tmp);

    my ($editor) =
        grep { $_ }
        map { $ENV{$_} }
        qw(SVN_EDITOR VISUAL EDITOR);
    $editor ||= 'vi';

    if (system($editor, $tmp->filename()) == -1) {
        croak "Can't start $editor\n";
    }

    open(my $in, "<", $tmp->filename())
        or croak "Can't read editor output\n";

    my $message = '';
    while (my $line = <$in>) {
        last if $line =~ /^$delimiter$/;
        $message .= $line;
    }
    close($in);
    chomp($message);

    croak "empty commit message, aborting" unless $message;

    return $message;
}

# get old changelog from package 

sub _get_old_package_changelog {
    my ($self, $package, %options) = @_;

    IO::String->require();

    my $handle =  IO::String->new();

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $self->{svn}->cat(
            $handle,
            sprintf(
                "%s/%s/log",
                $self->{config}->val('log', 'oldurl'),
                $package,
            ),
            $options{revision} || DEFAULT_REVISION,
            $pool
        ); 
    };
    if ($@) {
        if ($@ =~ /^Filesystem has no item/) {
            # no changelog for this package
        } else {
            croak "Can't get old changelog for $package: $@\n";
        }
    }

    seek($handle, 0, 0);

    my @cl;

    while(my $line = <$handle>) {
        chomp($line);
        $line or next;
        $line =~ /^%changelog/ and next;
        if ($line =~ /^\* (\w+\s+\w+\s+\d+\s+\d+)\s+(.*)/) {
            push(
                @cl,
                {
                    time   => str2time($1, 'UTC') || 0,
                    author => $2 || '',
                    text   => '',
                }
            );
        } else {
            if (!@cl) {
                push(@cl, { header => 1, text => '', 'time' => 0 });
            }
            $cl[-1]->{text} .= "$line\n";
        }
    }

    return @cl;
}

# get current changelog from package

sub _get_package_changelog {
    my ($self, $package, %options) = @_;

    my @entries;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    my $release_callback = sub {
        my ($changed_paths, $revision, $author, $date, $message) = @_;
        foreach my $path (keys %{$changed_paths || {}}) {
            my $info = $changed_paths->{$path};
            my $from_revision = $info->copyfrom_rev();
            next unless $from_revision > 0;
            next unless $path =~ m!releases/(?:([^:/]+):)?([^/]+)/([^/]+)!;
            push(
                @entries, {
                    revision => $from_revision,
                    author   => '',
                    time     => str2time($date),
                    text     => '',
                    evr      => $1 ? "$1:$2-$3" : "$2-$3",
                }
            );
        }
    };

    my $release_url =
        $self->get_package_url($package, %options, version => 'releases');
    if ($self->_check_url_exists($release_url)) {
        eval {
            $self->{svn}->log(
                $release_url,
                $options{revision} || DEFAULT_REVISION,
                0, 1, 0,
                $release_callback,
                $pool
            );
        };
        if ($@) {
            croak "Can't get svn log from release branch: $@\n";
        }
    }

    my $current_callback = sub {
        my ($changed_paths, $revision, $author, $date, $message) = @_;
        my $entry_to_update;
        foreach my $entry (@entries) {
            last if $revision > $entry->{revision};
            $entry_to_update = $entry;
        }

        if (!$entry_to_update) {
            $entry_to_update = {
                revision => $revision,
                author   => $author,
                time     => str2time($date),
                text     => '',
            };

            push(@entries, $entry_to_update);
        }

        $entry_to_update->{author} ||= $author;
        my @gti = gmtime(str2time($date));

        my ($textentry) =
            grep { $_->{author} eq $author }
            @{$entry_to_update->{changesets} || []};

        if (!$textentry) {
            $textentry = {
                author => $author,
                text => [],
            };
            push(@{$entry_to_update->{changesets}}, $textentry);
        }

        push(@{$textentry->{text}}, $message);
        
    };

    eval {
        $self->{svn}->log(
            $self->get_package_url($package, %options),
            $options{revision} || DEFAULT_REVISION,
            0, 0, 0,
            $current_callback,
            $pool
        );
    };
    if ($@) {
        croak "Can't get svn log from current branch: $@\n";
    }

    return @entries;
}

# format a changelog entry

sub _format_changelog_entry {
    my ($self, $entry) = @_;
    $entry->{time} or return $entry->{text};
    my @gti = gmtime($entry->{time});

    # subversion changelog is having empty commit
    return if not $entry->{author};
    my $text = $entry->{text};
    if (!$text) {
        my $indent = '';
        foreach my $changeset (@{$entry->{changesets} || []}) {
            if ($changeset->{author} ne $entry->{author}) {
                $text .=
                    "- from " .
                    $self->_get_formated_author($changeset->{author}) .
                    "\n";
            }
            foreach my $change (@{$changeset->{text}}) {
                next if $change =~ /SILENT/;
                my $dash = '- ';
                $change =~ s/^(\s|-|\*)*/$indent- /;
                foreach my $line (split(/\n/, $change)) {
                    next if $line =~ /SILENT/;
                    $line =~ s/^(\s|\*)*/$indent/;
                    if ($line !~ /^-/) {
                        $line =~ s/^/  /;
                    }
                    $text .= $line . "\n";
                    $dash = '  ';
                }
            }
            $indent = '  ';
        }
    }
    $text =~ s/^\*/-/gm;

    my $formated_entry;
    $formated_entry .= sprintf "* %s %s %s\n",
        strftime("%a %b %d %Y", @gti),
        $self->_get_formated_author($entry->{author}),
        $entry->{evr} || '';
    $formated_entry .= sprintf "+ Revision: %s\n", $entry->{revision}
        if $entry->{revision};
    $formated_entry .= sprintf "%s\n", $text;

    return $formated_entry;
}

# check given url exists

sub _check_url_exists {
    my ($self, $url, %options) = @_;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    eval {
        $self->{svn}->ls(
            $url, 
            $options{revision} || DEFAULT_REVISION,
            0,
            $pool
        );
    };
    if ($@) {
        return 0;
    }

    return 1;
}

# check given source exists

sub _check_source_exists {
    my ($self, $source) = @_;

    # create a new default pool for all local svn operations
    my $pool = SVN::Pool->new($self->{pool});

    my $text_status;
    $self->{svn}->status(
        $source,
        'HEAD',
        sub {
            my (undef, $status) = @_;
            $text_status = $status->text_status();
        },
        1, # recursive
        1, # get_all
        0, # update
        1, # no_ignore
    );

    return $text_status != $SVN::Wc::Status::unversioned;
}

sub _get_formated_author {
    my ($self, $author) = @_;

    # return cached result if it exists
    return $self->{authors}->{$author} if $self->{authors}->{$author};

    my $formated_author;
    my $server = $self->{config}->val('global', 'ldap-server');
    if ($server) {
        # lazy ldap connection
        if (!$self->{ldap}) {
            Net::LDAP->require();
            $self->{ldap} = Net::LDAP->new($server);
            $self->{ldap}->bind();
        }

        my $base = $self->{config}->val('global', 'ldap-base'),
        my $filter = $self->{config}->val('global', 'ldap-filterformat');
        $filter =~ s/\$username/$author/;

        my $result = $self->{ldap}->search(base => $base, filter => $filter);

        if ($result->count() == 0) { 
            $formated_author = $author;
        } else {
            my $entry = $result->entry(0);
            $formated_author = $self->{config}->val('global', 'ldap-resultformat');
            $formated_author =~ s/\$(\w+)/$entry->get_value($1)/eg;
        }
    } else {
        $formated_author = $self->{config}->val('users', $author, $author);
    }

    $self->{authors}->{$author} = $formated_author;
    return $formated_author;
}

1;
